/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * @(#)AccessBridgeJavaEntryPoints.h	1.15 03/09/02
 */

/* 
 * A class to manage JNI calls into AccessBridge.java
 */

#include "AccessBridgePackages.h"

#include <windows.h>
#include <jni.h>

#ifndef __AccessBridgeJavaEntryPoints_H__
#define __AccessBridgeJavaEntryPoints_H__

class AccessBridgeJavaEntryPoints {
	JNIEnv *jniEnv;

	jobject accessBridgeObject;

	jclass bridgeClass;
	jclass rectangleClass;
	jclass accessibleTextClass;

	jmethodID decrementReferenceMethod;
	jmethodID getJavaVersionPropertyMethod;
	jmethodID getAccessBridgeVersionMethod;

	jmethodID isJavaWindowMethod;
	jmethodID isSameObjectMethod;
	jmethodID getAccessibleContextFromHWNDMethod;

	jmethodID getAccessibleContextAtMethod;
	jmethodID getAccessibleContextWithFocusMethod;

	jmethodID getAccessibleNameFromContextMethod;
	jmethodID getAccessibleDescriptionFromContextMethod;
	jmethodID getAccessibleRoleStringFromContextMethod;
	jmethodID getAccessibleRoleStringFromContext_en_USMethod;
	jmethodID getAccessibleStatesStringFromContextMethod;
	jmethodID getAccessibleStatesStringFromContext_en_USMethod;
	jmethodID getAccessibleParentFromContextMethod;
	jmethodID getAccessibleIndexInParentFromContextMethod;
	jmethodID getAccessibleChildrenCountFromContextMethod;
	jmethodID getAccessibleChildFromContextMethod;
	jmethodID getAccessibleBoundsOnScreenFromContextMethod;
	jmethodID getAccessibleXcoordFromContextMethod;
	jmethodID getAccessibleYcoordFromContextMethod;
	jmethodID getAccessibleHeightFromContextMethod;
	jmethodID getAccessibleWidthFromContextMethod;

	jmethodID getAccessibleComponentFromContextMethod;
	jmethodID getAccessibleActionFromContextMethod;
	jmethodID getAccessibleSelectionFromContextMethod;
	jmethodID getAccessibleTextFromContextMethod;
	jmethodID getAccessibleValueFromContextMethod;

	/* begin AccessibleTable */
	jmethodID getAccessibleTableFromContextMethod;
	jmethodID getAccessibleTableRowHeaderMethod;
	jmethodID getAccessibleTableColumnHeaderMethod;
	jmethodID getAccessibleTableRowCountMethod;
	jmethodID getAccessibleTableColumnCountMethod;
	jmethodID getAccessibleTableCaptionMethod;
	jmethodID getAccessibleTableSummaryMethod;

	jmethodID getContextFromAccessibleTableMethod;
	jmethodID getAccessibleTableCellAccessibleContextMethod;
	jmethodID getAccessibleTableCellIndexMethod;
	jmethodID getAccessibleTableCellRowExtentMethod;
	jmethodID getAccessibleTableCellColumnExtentMethod;
	jmethodID isAccessibleTableCellSelectedMethod;

	jmethodID getAccessibleTableRowHeaderRowCountMethod;
	jmethodID getAccessibleTableColumnHeaderRowCountMethod;

	jmethodID getAccessibleTableRowHeaderColumnCountMethod;
	jmethodID getAccessibleTableColumnHeaderColumnCountMethod;

	jmethodID getAccessibleTableRowDescriptionMethod;
	jmethodID getAccessibleTableColumnDescriptionMethod;

	jmethodID getAccessibleTableRowSelectionCountMethod;
	jmethodID isAccessibleTableRowSelectedMethod;
	jmethodID getAccessibleTableRowSelectionsMethod;

	jmethodID getAccessibleTableColumnSelectionCountMethod;
	jmethodID isAccessibleTableColumnSelectedMethod;
	jmethodID getAccessibleTableColumnSelectionsMethod;

	jmethodID getAccessibleTableRowMethod;
	jmethodID getAccessibleTableColumnMethod;
	jmethodID getAccessibleTableIndexMethod;

	/* end AccessibleTable */

	/* begin AccessibleRelationSet */

	jmethodID getAccessibleRelationSetMethod;
	jmethodID getAccessibleRelationCountMethod;
	jmethodID getAccessibleRelationKeyMethod;
	jmethodID getAccessibleRelationTargetCountMethod;
	jmethodID getAccessibleRelationTargetMethod;

	/* end AccessibleRelationSet */

	jmethodID getAccessibleCharCountFromContextMethod;
	jmethodID getAccessibleCaretPositionFromContextMethod;
	jmethodID getAccessibleIndexAtPointFromContextMethod;
	
	jmethodID getAccessibleLetterAtIndexFromContextMethod;
	jmethodID getAccessibleWordAtIndexFromContextMethod;
	jmethodID getAccessibleSentenceAtIndexFromContextMethod;

	jmethodID getAccessibleTextSelectionStartFromContextMethod;
	jmethodID getAccessibleTextSelectionEndFromContextMethod;
	jmethodID getAccessibleTextSelectedTextFromContextMethod;
	jmethodID getAccessibleAttributesAtIndexFromContextMethod;
    jmethodID getAccessibleAttributeSetAtIndexFromContextMethod;
	jmethodID getAccessibleTextRectAtIndexFromContextMethod;
	jmethodID getAccessibleXcoordTextRectAtIndexFromContextMethod;
	jmethodID getAccessibleYcoordTextRectAtIndexFromContextMethod;
	jmethodID getAccessibleHeightTextRectAtIndexFromContextMethod;
	jmethodID getAccessibleWidthTextRectAtIndexFromContextMethod;
    jmethodID getAccessibleTextLineLeftBoundsFromContextMethod;
    jmethodID getAccessibleTextLineRightBoundsFromContextMethod;
    jmethodID getAccessibleTextRangeFromContextMethod;

    jmethodID getCurrentAccessibleValueFromContextMethod;
    jmethodID getMaximumAccessibleValueFromContextMethod;
    jmethodID getMinimumAccessibleValueFromContextMethod;

    jmethodID addAccessibleSelectionFromContextMethod;
    jmethodID clearAccessibleSelectionFromContextMethod;
    jmethodID getAccessibleSelectionContextFromContextMethod;
    jmethodID getAccessibleSelectionCountFromContextMethod;
    jmethodID isAccessibleChildSelectedFromContextMethod;
    jmethodID removeAccessibleSelectionFromContextMethod;
    jmethodID selectAllAccessibleSelectionFromContextMethod;

	jmethodID addJavaEventNotificationMethod;
	jmethodID removeJavaEventNotificationMethod;
	jmethodID addAccessibilityEventNotificationMethod;
	jmethodID removeAccessibilityEventNotificationMethod;

	jmethodID getBoldFromAttributeSetMethod;
	jmethodID getItalicFromAttributeSetMethod;
	jmethodID getUnderlineFromAttributeSetMethod;
	jmethodID getStrikethroughFromAttributeSetMethod;
	jmethodID getSuperscriptFromAttributeSetMethod;
	jmethodID getSubscriptFromAttributeSetMethod;
	jmethodID getBackgroundColorFromAttributeSetMethod;
	jmethodID getForegroundColorFromAttributeSetMethod;
	jmethodID getFontFamilyFromAttributeSetMethod;
	jmethodID getFontSizeFromAttributeSetMethod;
	jmethodID getAlignmentFromAttributeSetMethod;
	jmethodID getBidiLevelFromAttributeSetMethod;
	jmethodID getFirstLineIndentFromAttributeSetMethod;
	jmethodID getLeftIndentFromAttributeSetMethod;
	jmethodID getRightIndentFromAttributeSetMethod;
	jmethodID getLineSpacingFromAttributeSetMethod;
	jmethodID getSpaceAboveFromAttributeSetMethod;
	jmethodID getSpaceBelowFromAttributeSetMethod;


public:
    AccessBridgeJavaEntryPoints(JNIEnv *jniEnvironment, jobject bridgeObject);
    ~AccessBridgeJavaEntryPoints();
	BOOL BuildJavaEntryPoints();

	// HWND management methods
    BOOL isJavaWindow(jint window);
	BOOL isSameObject(jobject obj1, jobject obj2);
	jobject getAccessibleContextFromHWND(jint window);

    // version methods
	BOOL getVersionInfo(AccessBridgeVersionInfo *info);

	// verification methods
	BOOL verifyAccessibleText(jobject obj);

	// Accessible Context methods
	jobject getAccessibleContextAt(jint x, jint y, jobject AccessibleContext);
	jobject getAccessibleContextWithFocus();
	BOOL getAccessibleContextInfo(jobject AccessibleContext, AccessibleContextInfo *info);
	jobject getAccessibleChildFromContext(jobject AccessibleContext, jint childIndex);
	jobject getAccessibleParentFromContext(jobject AccessibleContext);

	/* begin AccessibleTable methods */

	BOOL getAccessibleTableInfo(jobject acParent, AccessibleTableInfo *tableInfo);
	BOOL getAccessibleTableCellInfo(jobject accessibleTable,jint row, jint column, 
		AccessibleTableCellInfo *tableCellInfo);

	BOOL getAccessibleTableRowHeader(jobject acParent, AccessibleTableInfo *tableInfo);
	BOOL getAccessibleTableColumnHeader(jobject acParent, AccessibleTableInfo *tableInfo);

	jobject getAccessibleTableRowDescription(jobject acParent, jint row);
	jobject getAccessibleTableColumnDescription(jobject acParent, jint column);

	jint getAccessibleTableRowSelectionCount(jobject accessibleTable);
	BOOL isAccessibleTableRowSelected(jobject accessibleTable, jint row);
	BOOL getAccessibleTableRowSelections(jobject accessibleTable, jint count, jint *selections);

	jint getAccessibleTableColumnSelectionCount(jobject accessibleTable);
	BOOL isAccessibleTableColumnSelected(jobject accessibleTable, jint column);
	BOOL getAccessibleTableColumnSelections(jobject accessibleTable, jint count, jint *selections);

	jint getAccessibleTableRow(jobject accessibleTable, jint index);
	jint getAccessibleTableColumn(jobject accessibleTable, jint index);
	jint getAccessibleTableIndex(jobject accessibleTable, jint row, jint column);

	/* end AccessibleTable methods */

	BOOL getAccessibleRelationSet(jobject accessibleContext, AccessibleRelationSetInfo *relationSetInfo);

	// Accessible Text methods
	BOOL getAccessibleTextInfo(jobject AccessibleContext, AccessibleTextInfo *textInfo, jint x, jint y);
	BOOL getAccessibleTextItems(jobject AccessibleContext, AccessibleTextItemsInfo *textItems, jint index);
	BOOL getAccessibleTextSelectionInfo(jobject AccessibleContext, AccessibleTextSelectionInfo *selectionInfo);
	BOOL getAccessibleTextAttributes(jobject AccessibleContext, jint index, AccessibleTextAttributesInfo *attributes);
	BOOL getAccessibleTextRect(jobject AccessibleContext, AccessibleTextRectInfo *rectInfo, jint index);
	BOOL getAccessibleTextLineBounds(jobject AccessibleContext, jint index, jint *startIndex, jint *endIndex);
	BOOL getAccessibleTextRange(jobject AccessibleContext, jint start, jint end, wchar_t *text, short len);

    // Accessible Value methods
    BOOL getCurrentAccessibleValueFromContext(jobject AccessibleContext, wchar_t *value, short len);
    BOOL getMaximumAccessibleValueFromContext(jobject AccessibleContext, wchar_t *value, short len);
    BOOL getMinimumAccessibleValueFromContext(jobject AccessibleContext, wchar_t *value, short len);

    // Accessible Selection methods
    void addAccessibleSelectionFromContext(jobject AccessibleContext, int i);
    void clearAccessibleSelectionFromContext(jobject AccessibleContext);
    jobject getAccessibleSelectionFromContext(jobject AccessibleContext, int i);
    int getAccessibleSelectionCountFromContext(jobject AccessibleContext);
    BOOL isAccessibleChildSelectedFromContext(jobject AccessibleContext, int i);
    void removeAccessibleSelectionFromContext(jobject AccessibleContext, int i);
    void selectAllAccessibleSelectionFromContext(jobject AccessibleContext);

	// Event handling methods
	BOOL addJavaEventNotification(jlong type);
	BOOL removeJavaEventNotification(jlong type);
	BOOL addAccessibilityEventNotification(jlong type);
	BOOL removeAccessibilityEventNotification(jlong type);

};

#endif
