/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * @(#)WinAccessBridge.cpp	1.17 03/09/02
 */

/* 
 * A DLL which is loaded by Windows executables to handle communication
 * between Java VMs purposes of Accessbility.
 */

#ifdef _DEBUG
	#define DEBUGGING_ON
#endif


#include "AccessBridgeDebug.h"
#include "WinAccessBridge.h"
#include "accessBridgeResource.h"
#include "accessBridgeCallbacks.h"
#include "AccessBridgeMessages.h"
#include "AccessBridgeMessageQueue.h"

#include <windows.h>
#include <jni.h>
#include <stdio.h>
//#include <javaString.h>


WinAccessBridge *theWindowsAccessBridge;
HWND theDialogWindow;

// unique broadcast msg. IDs gotten dymanically
extern UINT theFromJavaHelloMsgID;
extern UINT theFromWindowsHelloMsgID;


// ---------------------------------------------------------------------------

extern "C" {
	/**	
	 * DllMain - where Windows executables will load/unload us
	 *
	 */
	BOOL WINAPI DllMain(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved) {

		switch (fdwReason) {
		case DLL_PROCESS_ATTACH: 	// A Windows executable loaded us
			theWindowsAccessBridge = new WinAccessBridge(hinstDll);
			break;

		case DLL_PROCESS_DETACH:	// A Windows executable unloaded us
			if (theWindowsAccessBridge != (WinAccessBridge *) 0) {
				PrintDebugString("*** AccessBridgeDialogProc -> deleting theWindowsAccessBridge");
				delete theWindowsAccessBridge;
			}
			break; 
		}

/*
        DEBUG(int result);
        DEBUG(OFSTRUCT fileStruct);
        DEBUG(HFILE debugFile = OpenFile("\\AccessBridgeDebugging.txt", &fileStruct, OF_CREATE | OF_READWRITE));
        DEBUG(result = _CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_FILE | _CRTDBG_MODE_DEBUG));
        DEBUG(_CrtSetReportFile(_CRT_WARN, debugFile));
*/
		return(TRUE);
	}

	/**
	 * Append debug info to dialog
	 *
	 *
	void AppendToCallInfo(char *s) {
		char buffer[4096];

		GetDlgItemText(theDialogWindow, cCallInfo, buffer, sizeof(buffer));
		if (strlen(buffer) < (sizeof(buffer) - strlen(s))) {
			strncat(buffer, s, sizeof(buffer));
			SetDlgItemText(theDialogWindow, cCallInfo, buffer);
		} else {
			SetDlgItemText(theDialogWindow, cCallInfo, s);
		}
	}
    */
    

	/**
	 * Append debug info to dialog
     *
     * replaced with code to send output to debug file
	 *
	 */
	void AppendToCallInfo(char *s) {

/*
        _CrtDbgReport(_CRT_WARN, (const char *) NULL, NULL, (const char *) NULL,
                      (const char *) "WinAccessBridge: %s", s);
*/

        char buf[1024];
        sprintf(buf, "WinAccessBridge: %s", s);
        OutputDebugString(buf);
	}
//    */ 

	/**
	 * Our window proc
	 *
	 */
	BOOL CALLBACK AccessBridgeDialogProc(HWND hDlg, UINT message, UINT wParam, LONG lParam) {
		COPYDATASTRUCT *sentToUs;
		char *package;

		switch (message) {
		case WM_INITDIALOG:
			PrintDebugString("AccessBridgeDialogProc -> Initializing");
			break;
		
			// call from Java with data for us to deliver
		case WM_COPYDATA:
			if (theDialogWindow == (HWND) wParam) {
				PrintDebugString("\r\nAccessBridgeDialogProc -> Got WM_COPYDATA from Java Bridge DLL");
			} else {
				PrintDebugString("\r\nAccessBridgeDialogProc -> Got WM_COPYDATA from HWND %x", wParam); 
				sentToUs = (COPYDATASTRUCT *) lParam;
				package = (char *) sentToUs->lpData;
				theWindowsAccessBridge->preProcessPackage(package, sentToUs->cbData);
			}
			break;

			// message to ourselves -> de-queue messages and send 'em
		case AB_MESSAGE_QUEUED:
			PrintDebugString("\r\nAccessBridgeDialogProc -> Got AB_MESSAGE_QUEUED from ourselves");
			theWindowsAccessBridge->receiveAQueuedPackage();
			break;

			// a JavaAccessBridge DLL is going away
		case AB_DLL_GOING_AWAY:
			PrintDebugString("\r\n***** AccessBridgeDialogProc -> Got AB_DLL_GOING_AWAY message");
			theWindowsAccessBridge->JavaVMDestroyed((HWND) wParam);
			break;

		default:
			// the JavaVM is saying "hi"!
			// wParam == sourceHwnd; lParam == JavaVMID
			if (message == theFromJavaHelloMsgID) {
				PrintDebugString("\r\nAccessBridgeDialogProc -> Got theFromJavaHelloMsgID; wParam = %X, lParam = %X", wParam, lParam);
				theWindowsAccessBridge->rendezvousWithNewJavaDLL((HWND) wParam, (long ) lParam);
			}
			break;
		}

		return (FALSE);   
	}

}

// ---------------------------------------------------------------------------

/**
 * Initialize the WinAccessBridge
 *
 */
WinAccessBridge::WinAccessBridge(HINSTANCE hInstance) {
	windowsInstance = hInstance;
	javaVMs = (AccessBridgeJavaVMInstance *) 0;
	eventHandler = new AccessBridgeEventHandler();
	messageQueue = new AccessBridgeMessageQueue();
	initBroadcastMessageIDs();		// get the unique to us broadcast msg. IDs
}

/**
 * Destroy the WinAccessBridge
 *
 */
WinAccessBridge::~WinAccessBridge() {
	// inform all other AccessBridges that we're going away
	//  -> shut down all event listening
	//  -> release all objects held in the JVM by us

	PrintDebugString("\r\n*****in WinAccessBridge::~WinAccessBridge()"); 

	// send a broadcast msg.; let other AccessBridge DLLs know we're going away
	AccessBridgeJavaVMInstance *current = javaVMs;
	while (current != (AccessBridgeJavaVMInstance *) 0) {
		PrintDebugString("  telling %X we're going away", current->javaAccessBridgeWindow); 
		SendMessage(current->javaAccessBridgeWindow,
					AB_DLL_GOING_AWAY, (WPARAM) dialogWindow, (LPARAM) 0);
		current = current->nextJVMInstance;
	}

	PrintDebugString("  finished telling JVMs about our demise"); 

	delete eventHandler;
	delete messageQueue;
	delete javaVMs;

	PrintDebugString("  finished deleting eventHandler, messageQueue, and javaVMs"); 
	PrintDebugString("GOODBYE CRUEL WORLD..."); 

    DestroyWindow(theDialogWindow);
}


/**
 * Bring up our window; make a connection to the rest of the world
 *
 */
BOOL
WinAccessBridge::initWindow() {
	theDialogWindow = CreateDialog(windowsInstance,
								   "ACCESSBRIDGESTATUSWINDOW", NULL,
								   (DLGPROC) AccessBridgeDialogProc);

	// If window could not be created, return "failure".
	if (!theDialogWindow)
	   return (FALSE);

	dialogWindow = theDialogWindow;

	// Make the window visible, update its client area, & return "success".
	DEBUG(ShowWindow (theDialogWindow, SW_SHOWNORMAL));
	DEBUG(UpdateWindow (theDialogWindow)); 

	// post a broadcast msg.; let other AccessBridge DLLs know we exist
	PostMessage(HWND_BROADCAST, theFromWindowsHelloMsgID, (WPARAM) dialogWindow, (LPARAM) 0);
	
	return (TRUE);      
}

// -----------------------

/**
 * rendezvousWithNewJavaDLL
 *		- Build AccessBridgeJavaVMInstance data structure
 *		  (including setting up Memory-Mapped file info)
 *
 */
LRESULT
WinAccessBridge::rendezvousWithNewJavaDLL(HWND JavaBridgeDLLwindow, long vmID) {
	LRESULT returnVal;

	PrintDebugString("\r\nin JavaAccessBridge::rendezvousWithNewJavaDLL(%X, %X)",
		  JavaBridgeDLLwindow, vmID);

	AccessBridgeJavaVMInstance *newVM = 
		new AccessBridgeJavaVMInstance(dialogWindow, JavaBridgeDLLwindow, vmID, javaVMs);
	javaVMs = newVM;

	returnVal = javaVMs->initiateIPC();
	if (returnVal == 0) {

		// tell the newly created JavaVM what events we're interested in, if any
		long javaEventMask = eventHandler->getJavaEventMask();
		long accessibilityEventMask = eventHandler->getAccessibilityEventMask();

		PrintDebugString("  Setting Java event mask to: %X", javaEventMask);

		if (javaEventMask != 0) {
			addJavaEventNotification(javaEventMask);
		}

		PrintDebugString("  Setting Accessibility event mask to: %X", accessibilityEventMask);

		if (accessibilityEventMask != 0) {
			addAccessibilityEventNotification(accessibilityEventMask);
		}
    } else {
		PrintDebugString("  ERROR: Failed to initiate IPC with newly created JavaVM!!!");
		return FALSE;
	}

	PrintDebugString("  Success!!  We rendezvoused with the JavaDLL");
	return returnVal;
}

// -----------------------

/**
 * sendPackage - uses SendMessage(WM_COPYDATA) to do IPC messaging 
 *				 with the Java AccessBridge DLL
 *
 *				 NOTE: WM_COPYDATA is only for one-way IPC; there
 *				 is now way to return parameters (especially big ones)
 *				 Use sendMemoryPackage() to do that!
 */
void
WinAccessBridge::sendPackage(char *buffer, long bufsize, HWND destWindow) {
	COPYDATASTRUCT toCopy;
	toCopy.dwData = 0;		// 32-bits we could use for something...
	toCopy.cbData = bufsize;
	toCopy.lpData = buffer;

	SendMessage(destWindow, WM_COPYDATA, (WPARAM) dialogWindow, (LPARAM) &toCopy);
}


/**
 * sendMemoryPackage - uses Memory-Mapped files to do IPC messaging 
 *					   with the Java AccessBridge DLL, informing the
 *					   Java AccessBridge DLL via SendMessage that something
 *					   is waiting for it in the shared file...
 *
 *					   In the SendMessage call, the third param (WPARAM) is
 *					   the source HWND (theDialogWindow in this case), and
 *					   the fourth param (LPARAM) is the size in bytes of
 *					   the package put into shared memory.
 *
 */
BOOL
WinAccessBridge::sendMemoryPackage(char *buffer, long bufsize, HWND destWindow) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	AccessBridgeJavaVMInstance *ourABJavaVMInstance;
	ourABJavaVMInstance = javaVMs->findABJavaVMInstanceFromJavaHWND(destWindow);
	if (ourABJavaVMInstance != (AccessBridgeJavaVMInstance *) 0) {
		ourABJavaVMInstance->sendMemoryPackage(buffer, bufsize);
		return TRUE;
	} else {
		PrintDebugString("ERROR sending memory package: couldn't find destWindow");
		return FALSE;
	}
}


/**
 * queuePackage - put a package onto the queue for latter processing
 *
 */
BOOL
WinAccessBridge::queuePackage(char *buffer, long bufsize) {
	PrintDebugString("  in WinAccessBridge::queuePackage(%X, %d)", buffer, bufsize); 

	AccessBridgeQueueElement *element = new AccessBridgeQueueElement(buffer, bufsize);

	messageQueue->add(element);
	PostMessage(dialogWindow, AB_MESSAGE_QUEUED, (WPARAM) 0, (LPARAM) 0);
	return TRUE;
}


/**
 * receiveAQueuedPackage - remove a pending packge from the queue and
 *						   handle it. If the queue is busy, post a
 *						   message to self to retrieve it later
 *
 */
BOOL
WinAccessBridge::receiveAQueuedPackage() {
	AccessBridgeQueueElement *element;

	PrintDebugString("\r\nin WinAccessBridge::receiveAQueuedPackage()");

    // ensure against re-entrancy problems...
    if (messageQueue->getRemoveLockSetting() == FALSE) {
        messageQueue->setRemoveLock(TRUE);

       	PrintDebugString("  dequeueing message");

	    QueueReturns result = messageQueue->remove(&element);

	    PrintDebugString("   'element->buffer' contains:");
	    DEBUG(PackageType *type = (PackageType *) element->buffer); 
	    DEBUG(FocusGainedPackageTag *pkg = (FocusGainedPackageTag *) (((char *) element->buffer) + sizeof(PackageType))); 
	    DEBUG(PrintDebugString("     PackageType = %X", *type)); 
	    DEBUG(PrintDebugString("     EventPackage: vmID = %X, event = %X, source = %X", pkg->vmID, pkg->Event, pkg->AccessibleContextSource)); 

	    switch (result) {

	    case cQueueBroken:
		    PrintDebugString("  ERROR!!! Queue seems to be broken!");
            messageQueue->setRemoveLock(FALSE);
		    return FALSE;

	    case cMoreMessages:
	    case cQueueEmpty:
		    if (element != (AccessBridgeQueueElement *) 0) {
			    PrintDebugString("  found one; sending it!");
			    processPackage(element->buffer, element->bufsize);
			    delete element;
		    } else {
			    PrintDebugString("  ODD... element == 0!");
			    return FALSE;
		    }
		    break;

	    case cQueueInUse:
		    PrintDebugString("  Queue in use, will try again later...");
		    PostMessage(dialogWindow, AB_MESSAGE_QUEUED, (WPARAM) 0, (LPARAM) 0);
		    break;

	    default:
            messageQueue->setRemoveLock(FALSE);
		    return FALSE;	// should never get something we don't recognize!
	    }
    } else {
       	PrintDebugString("  unable to dequeue message; remove lock is set");
    }

    messageQueue->setRemoveLock(FALSE);
	return TRUE;
}

// -----------------------

/**
 * preProcessPackage 
 *		- do triage on incoming packages; queue some, deal with others
 *
 */
void
WinAccessBridge::preProcessPackage(char *buffer, long bufsize) {
	PrintDebugString("\r\nPreProcessing package sent from Java:");

	PackageType *type = (PackageType *) buffer;

	switch (*type) {

		// event packages all get queued for later handling
//	case cPropertyChangePackage:
	case cFocusGainedPackage:
	case cFocusLostPackage:
	case cCaretUpdatePackage:
	case cMouseClickedPackage:
	case cMouseEnteredPackage:
	case cMouseExitedPackage:
	case cMousePressedPackage:
	case cMouseReleasedPackage:
	case cMenuCanceledPackage:
	case cMenuDeselectedPackage:
	case cMenuSelectedPackage:
	case cPopupMenuCanceledPackage:
	case cPopupMenuWillBecomeInvisiblePackage:
	case cPopupMenuWillBecomeVisiblePackage:

   	case cPropertyCaretChangePackage:
   	case cPropertyDescriptionChangePackage:
   	case cPropertyNameChangePackage:
   	case cPropertySelectionChangePackage:
   	case cPropertyStateChangePackage:
   	case cPropertyTextChangePackage:
   	case cPropertyValueChangePackage:
   	case cPropertyVisibleDataChangePackage:
   	case cPropertyChildChangePackage:
   	case cPropertyActiveDescendentChangePackage:

	case cPropertyTableModelChangePackage:

        queuePackage(buffer, bufsize);
        break;

	// perhaps there will be some other packages to process at some point... //

	default:
		PrintDebugString("   processing FAILED!! -> don't know how to handle type = %X", *type);
		break;
	}

	PrintDebugString("   package preprocessing completed");
}


#define DISPATCH_EVENT_PACKAGE(packageID, eventPackage, fireEventMethod)			\
	case packageID:																	\
		PrintDebugString("   type == <some type>");				                \
		if (bufsize == sizeof(PackageType) + sizeof(eventPackage)) {				\
			eventPackage *pkg =														\
				(eventPackage *) (buffer + sizeof(PackageType));					\
			theWindowsAccessBridge->eventHandler->fireEventMethod(					\
				pkg->vmID, pkg->Event, pkg->AccessibleContextSource);				\
			PrintDebugString("   call successful");			                    \
			PrintDebugString("   event fired!");					                \
		} else {																	\
			PrintDebugString("   processing FAILED!! -> bufsize = %d; expectation = %d", \
				  bufsize, sizeof(PackageType) + sizeof(eventPackage));			\
		}																			\
		break;

#define DISPATCH_PROPERTY_CHANGE_PACKAGE(packageID, eventPackage, fireEventMethod, oldValue, newValue) \
	case packageID:                                                                 \
		PrintDebugString("   type == <some type>");				                \
		if (bufsize == sizeof(PackageType) + sizeof(eventPackage)) {				\
			eventPackage *pkg =														\
				(eventPackage *) (buffer + sizeof(PackageType));					\
			theWindowsAccessBridge->eventHandler->fireEventMethod(					\
				pkg->vmID, pkg->Event, pkg->AccessibleContextSource,				\
				pkg->oldValue, pkg->newValue);                                      \
			PrintDebugString("   call successful");			                    \
			PrintDebugString("   event fired!");					                \
		} else {																	\
			PrintDebugString("   processing FAILED!! -> bufsize = %d; expectation = %d", \
				  bufsize, sizeof(PackageType) + sizeof(eventPackage));			\
		}																			\
		break;

#define DISPATCH_PROPERTY_TABLE_MODEL_CHANGE_PACKAGE(packageID, eventPackage, fireEventMethod, oldValue, newValue) \
	case packageID:                                                                 \
		PrintDebugString("   type == <some type>");				                \
		if (bufsize == sizeof(PackageType) + sizeof(eventPackage)) {				\
			eventPackage *pkg =														\
				(eventPackage *) (buffer + sizeof(PackageType));					\
			theWindowsAccessBridge->eventHandler->fireEventMethod(					\
				pkg->vmID, pkg->Event, pkg->AccessibleContextSource,				\
				pkg->oldValue, pkg->newValue);                                      \
			PrintDebugString("   call successful");			                    \
			PrintDebugString("   event fired!");					                \
		} else {																	\
			PrintDebugString("   processing FAILED!! -> bufsize = %d; expectation = %d", \
				  bufsize, sizeof(PackageType) + sizeof(eventPackage));			\
		}																			\
		break;

/**
 * processPackage - processes the output of SendMessage(WM_COPYDATA) 
 *					to do IPC messaging with the Java AccessBridge DLL
 *
 */
void
WinAccessBridge::processPackage(char *buffer, long bufsize) {
	PrintDebugString("\r\nWinAccessBridge::Processing package sent from Java:");

	PackageType *type = (PackageType *) buffer;

	switch (*type) {

	DISPATCH_EVENT_PACKAGE(cFocusGainedPackage, FocusGainedPackage, fireFocusGained);
	DISPATCH_EVENT_PACKAGE(cFocusLostPackage, FocusLostPackage, fireFocusLost);

	DISPATCH_EVENT_PACKAGE(cCaretUpdatePackage, CaretUpdatePackage, fireCaretUpdate);

	DISPATCH_EVENT_PACKAGE(cMouseClickedPackage, MouseClickedPackage, fireMouseClicked);
	DISPATCH_EVENT_PACKAGE(cMouseEnteredPackage, MouseEnteredPackage, fireMouseEntered);
	DISPATCH_EVENT_PACKAGE(cMouseExitedPackage, MouseExitedPackage, fireMouseExited);
	DISPATCH_EVENT_PACKAGE(cMousePressedPackage, MousePressedPackage, fireMousePressed);
	DISPATCH_EVENT_PACKAGE(cMouseReleasedPackage, MouseReleasedPackage, fireMouseReleased);

	DISPATCH_EVENT_PACKAGE(cMenuCanceledPackage, MenuCanceledPackage, fireMenuCanceled);
	DISPATCH_EVENT_PACKAGE(cMenuDeselectedPackage, MenuDeselectedPackage, fireMenuDeselected);
	DISPATCH_EVENT_PACKAGE(cMenuSelectedPackage, MenuSelectedPackage, fireMenuSelected);
	DISPATCH_EVENT_PACKAGE(cPopupMenuCanceledPackage, PopupMenuCanceledPackage, firePopupMenuCanceled);
	DISPATCH_EVENT_PACKAGE(cPopupMenuWillBecomeInvisiblePackage, PopupMenuWillBecomeInvisiblePackage, firePopupMenuWillBecomeInvisible);
	DISPATCH_EVENT_PACKAGE(cPopupMenuWillBecomeVisiblePackage, PopupMenuWillBecomeVisiblePackage, firePopupMenuWillBecomeVisible);
  
    DISPATCH_PROPERTY_CHANGE_PACKAGE(cPropertyNameChangePackage, 
                                     PropertyNameChangePackage, 
                                     firePropertyNameChange, oldName, newName)
    DISPATCH_PROPERTY_CHANGE_PACKAGE(cPropertyDescriptionChangePackage, 
                                     PropertyDescriptionChangePackage, 
                                     firePropertyDescriptionChange,
                                     oldDescription, newDescription)
    DISPATCH_PROPERTY_CHANGE_PACKAGE(cPropertyStateChangePackage, 
                                     PropertyStateChangePackage, 
                                     firePropertyStateChange, oldState, newState)
    DISPATCH_PROPERTY_CHANGE_PACKAGE(cPropertyValueChangePackage, 
                                     PropertyValueChangePackage, 
                                     firePropertyValueChange, oldValue, newValue)
    DISPATCH_EVENT_PACKAGE(cPropertySelectionChangePackage, 
                           PropertySelectionChangePackage, firePropertySelectionChange)
    DISPATCH_EVENT_PACKAGE(cPropertyTextChangePackage, 
                           PropertyTextChangePackage, firePropertyTextChange)
    DISPATCH_PROPERTY_CHANGE_PACKAGE(cPropertyCaretChangePackage, 
                                     PropertyCaretChangePackage, 
                                     firePropertyCaretChange, oldPosition, newPosition)
    DISPATCH_EVENT_PACKAGE(cPropertyVisibleDataChangePackage, 
                           PropertyVisibleDataChangePackage, firePropertyVisibleDataChange)
    DISPATCH_PROPERTY_CHANGE_PACKAGE(cPropertyChildChangePackage, 
                                     PropertyChildChangePackage, 
                                     firePropertyChildChange, 
                                     oldChildAccessibleContext, 
                                     newChildAccessibleContext)
    DISPATCH_PROPERTY_CHANGE_PACKAGE(cPropertyActiveDescendentChangePackage, 
                                     PropertyActiveDescendentChangePackage, 
                                     firePropertyActiveDescendentChange, 
                                     oldActiveDescendentAccessibleContext, 
                                     newActiveDescendentAccessibleContext)

	DISPATCH_PROPERTY_TABLE_MODEL_CHANGE_PACKAGE(cPropertyTableModelChangePackage, 
                                     PropertyTableModelChangePackage, 
                                     firePropertyTableModelChange, 
									 oldValue, newValue)


	default:
		PrintDebugString("   processing FAILED!! -> don't know how to handle type = %X", *type);
		break;
	}

	PrintDebugString("   package processing completed");
}


// -----------------------------

void
WinAccessBridge::JavaVMDestroyed(HWND VMBridgeDLLWindow) {
	PrintDebugString("\r\n***** WinAccessBridge::JavaVMDestroyed(%X)", VMBridgeDLLWindow);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}

	AccessBridgeJavaVMInstance *currentVM = javaVMs;
	AccessBridgeJavaVMInstance *previousVM = javaVMs;
	if (javaVMs->javaAccessBridgeWindow == VMBridgeDLLWindow) {
		javaVMs = javaVMs->nextJVMInstance;
		delete currentVM;

		PrintDebugString("  data structures successfully removed");

		// [[[FIXME]]] inform Windows AT that a JVM went away,
		// and that any jobjects it's got lying around for that JVM
		// are now invalid

	} else {
		while (currentVM != (AccessBridgeJavaVMInstance *) 0) {
			if (currentVM->javaAccessBridgeWindow == VMBridgeDLLWindow) {
				previousVM->nextJVMInstance = currentVM->nextJVMInstance;
				delete currentVM;

				PrintDebugString("  data structures successfully removed");

				// [[[FIXME]]] inform Windows AT that a JVM went away,
				// and that any jobjects it's got lying around for that JVM
				// are now invalid

				return;
			} else {
				previousVM = currentVM;
				currentVM = currentVM->nextJVMInstance;
			}
		}
		PrintDebugString("  ERROR!! couldn't find matching data structures!");
	}
}

// -----------------------

/**
 * releaseJavaObject - lets the JavaVM know it can release the Java Object
 *
 * Note: once you have made this call, the JavaVM will garbage collect
 * the jobject you pass in.  If you later use that jobject in another
 * call, you will cause all maner of havoc!
 *
 */
void
WinAccessBridge::releaseJavaObject(long vmID, jobject object) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(ReleaseJavaObjectPackage)];
	PackageType *type = (PackageType *) buffer;
	ReleaseJavaObjectPackage *pkg = (ReleaseJavaObjectPackage *) (buffer + sizeof(PackageType));
	*type = cReleaseJavaObjectPackage;
	pkg->vmID = vmID;
	pkg->object = object;

	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		sendPackage(buffer, sizeof(buffer), destABWindow);		// no return values!
	}
}

// -----------------------

/**
 * getVersionInfo - fill the AccessBridgeVersionInfo struct
 *
 */
BOOL
WinAccessBridge::getVersionInfo(long vmID, AccessBridgeVersionInfo *info) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessBridgeVersionPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessBridgeVersionPackage *pkg = (GetAccessBridgeVersionPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessBridgeVersionPackage;
	pkg->vmID = vmID;

    PrintDebugString("\r\nWinAccessBridge::getVersionInfo(%X, )", vmID);
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
        if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
	        memcpy(info, &(pkg->rVersionInfo), sizeof(AccessBridgeVersionInfo));
            wcsncpy(info->bridgeWinDLLVersion, 
                    WINDOW_ACCESS_BRIDGE_VERSION,
                    sizeof(info->bridgeWinDLLVersion) / sizeof(wchar_t));
	        return TRUE;
        }
	}
    wcsncpy(info->bridgeWinDLLVersion, 
            WINDOW_ACCESS_BRIDGE_VERSION,
            sizeof(info->bridgeWinDLLVersion) / sizeof(wchar_t));
    return FALSE;
}


/********** Window-related routines ***********************************/

/**
 * isJavaWindow - returns TRUE if the HWND is a top-level Java Window
 *
 * Note: just because the Windnow is a top-level Java window, that doesn't 
 * mean that it is accessible.  Call getAccessibleContextFromHWND(HWND) to get the 
 * AccessibleContext, if any, for an HWND that is a Java Window.
 *
 */
BOOL
WinAccessBridge::isJavaWindow(HWND window) {
	HWND hwnd;

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}

    // quick check to see if 'window' is top-level; if not, it's not interesting...
    // [[[FIXME]]] is this for sure an OK optimization?
	hwnd = getTopLevelHWND(window);
	if (hwnd == (HWND) NULL) {
		return FALSE;
	}

	PrintDebugString("  in WinAccessBridge::isJavaWindow");

    char buffer[sizeof(PackageType) + sizeof(IsJavaWindowPackage)];
	PackageType *type = (PackageType *) buffer;
	IsJavaWindowPackage *pkg = (IsJavaWindowPackage *) (buffer + sizeof(PackageType));
	*type = cIsJavaWindowPackage;
	pkg->window = (jint) window;

    PrintDebugString("\r\nWinAccessBridge::isJavaWindow(%X)", window);

	AccessBridgeJavaVMInstance *current = javaVMs;
	while (current != (AccessBridgeJavaVMInstance *) 0) {

		if (sendMemoryPackage(buffer, sizeof(buffer), current->javaAccessBridgeWindow) == TRUE) {
			if (pkg->rResult != 0) {
				return TRUE;
			}
		}
		current = current->nextJVMInstance;
	}

	return FALSE;
    

/*
	char classname[256];
	HWND hwnd;

	hwnd = getTopLevelHWND(window);
	if (hwnd == (HWND) NULL) {
		return FALSE;
	}
	GetClassName(hwnd, classname, 256);
	
	if (strstr(classname, "AwtFrame") != 0) {
		return TRUE;
	} else if (strstr(classname, "AwtWindow") != 0) {
		return TRUE;
	} else if (strstr(classname, "AwtDialog") != 0) {
		return TRUE;
    }
*/
    // JDK 1.4 introduces new (and changes old) classnames
/*
    else if (strstr(classname, "SunAwtToolkit") != 0) {
        return TRUE;
    } else if (strstr(classname, "javax.swing.JFrame") != 0) {
        return TRUE;
    }
*/

	return FALSE;
}


/**
 * isSameObject - returns TRUE if the two object references refer to
 *     the same object. Otherwise, this method returns FALSE:
 */
BOOL
WinAccessBridge::isSameObject(long vmID, jobject obj1, jobject obj2) {

	PrintDebugString("\r\nWinAccessBridge::isSameObject(%X %X)", obj1, obj2);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}

    char buffer[sizeof(PackageType) + sizeof(IsSameObjectPackage)];
	PackageType *type = (PackageType *) buffer;
	IsSameObjectPackage *pkg = (IsSameObjectPackage *) (buffer + sizeof(PackageType));
	*type = cIsSameObjectPackage;
	pkg->vmID = vmID;
	pkg->obj1 = obj1;
	pkg->obj2 = obj2;

	HWND destABWindow = javaVMs->findAccessBridgeWindow(pkg->vmID);
	if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
		if (pkg->rResult != 0) {
			PrintDebugString("  WinAccessBridge::isSameObject returning TRUE (same object)");
			return TRUE;
		} else {
			PrintDebugString("  WinAccessBridge::isSameObject returning FALSE (di");
			return FALSE;
		}
	}
	PrintDebugString("  WinAccessBridge::isSameObject returning FALSE (sendMemoryPackage failed)");
	return FALSE;
}


/**
 * FromHWND - returns the AccessibleContext jobject for the HWND
 *
 * Note: this routine can return null, even if the HWND is a Java Window,
 * because the Java Window may not be accessible.
 *
 */
BOOL
WinAccessBridge::getAccessibleContextFromHWND(HWND window, long *vmID, jobject *AccessibleContext) {
	PrintDebugString("  in WinAccessBridge::getAccessibleContextFromHWND");
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}

    char buffer[sizeof(PackageType) + sizeof(GetAccessibleContextFromHWNDPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleContextFromHWNDPackage *pkg = (GetAccessibleContextFromHWNDPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleContextFromHWNDPackage;
	pkg->window = (jint) window;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleContextFromHWND(%X, %X, )", window, vmID);

    DEBUG(pkg->rVMID = (long ) 0x01010101);
	DEBUG(pkg->rAccessibleContext = (jobject) 0x01010101);

	AccessBridgeJavaVMInstance *current = javaVMs;
	while (current != (AccessBridgeJavaVMInstance *) 0) {

		if (sendMemoryPackage(buffer, sizeof(buffer), current->javaAccessBridgeWindow) == TRUE) {
			if (pkg->rAccessibleContext != 0) {
				*vmID = pkg->rVMID;
				*AccessibleContext = pkg->rAccessibleContext;
				if (pkg->rVMID != current->vmID) {
					PrintDebugString("   ERROR! getAccessibleContextFromHWND vmIDs don't match!");
					PrintDebugString("   -> pkg->rVMID = %X", pkg->rVMID);
					PrintDebugString("   ->  current->vmID = %X", current->vmID);
					PrintDebugString("   -> pkg->rAccessibleContext = %X", pkg->rAccessibleContext);
					return FALSE;
				}
				return TRUE;
			}
		}
		current = current->nextJVMInstance;
	}

	return FALSE;
}



/********** AccessibleContext routines ***********************************/

/**
 * Walk through Java Windows, in front-to-back Z-order.
 * If NULL is passed it, this function starts at the top.
 *
 */
HWND
WinAccessBridge::getNextJavaWindow(HWND previous) {
	HWND current = previous;
	if (current == NULL) {
		current = GetTopWindow(NULL);
	} else {
		current = GetNextWindow(current, GW_HWNDNEXT);
	}
	while (current != NULL) {
		if (isJavaWindow(current)) {
			return current;
		}
		current = GetNextWindow(current, GW_HWNDNEXT);
	}
	return NULL;
}


/**
 * getAccessibleContextAt - performs the Java code:
 *   Accessible a = EventQueueMonitor.getAccessibleAt(x, y);
 *	 return a.getAccessibleContext();
 *
 * Note: this call explicitly goes through the AccessBridge,
 * so that the AccessBridge can hide expected changes in how this functions 
 * between JDK 1.1.x w/AccessibilityUtility classes, and JDK 1.2, when some 
 * of this functionality may be built into the platform
 *
 */
BOOL
WinAccessBridge::getAccessibleContextAt(long vmID, jobject AccessibleContextParent, 
										jint x, jint y, jobject *AccessibleContext) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}

	char buffer[sizeof(PackageType) + sizeof(GetAccessibleContextAtPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleContextAtPackage *pkg = (GetAccessibleContextAtPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleContextAtPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContextParent;
	pkg->x = x;
	pkg->y = y;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleContextAt(%X, %X, %d, %c)", vmID, AccessibleContextParent, x, y);
	HWND destABWindow = javaVMs->findAccessBridgeWindow(pkg->vmID);
	if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
		*AccessibleContext = pkg->rAccessibleContext;
		return TRUE;
	}

	return FALSE;
}


/**
 * getAccessibleContextWithFocus - performs the Java code:
 *   Accessible a = Translator.getAccessible(SwingEventMonitor.getComponentWithFocus());
 *   return a.getAccessibleContext();
 *
 * Note: this call explicitly goes through the AccessBridge,
 * so that the AccessBridge can hide expected changes in how this functions 
 * between JDK 1.1.x w/AccessibilityUtility classes, and JDK 1.2, when some 
 * of this functionality may be built into the platform
 *
 */
BOOL
WinAccessBridge::getAccessibleContextWithFocus(HWND window, long *vmID, jobject *AccessibleContext) {

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleContextWithFocusPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleContextWithFocusPackage *pkg = (GetAccessibleContextWithFocusPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleContextWithFocusPackage;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleContextWithFocus(%X, %X, )", window, vmID);
		// find vmID, etc. from HWND; ask that VM for the AC w/Focus
	if (getAccessibleContextFromHWND(window, &(pkg->rVMID), &(pkg->rAccessibleContext)) == TRUE) {
		HWND destABWindow = javaVMs->findAccessBridgeWindow(pkg->rVMID);	// ineffecient [[[FIXME]]]
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			*vmID = pkg->rVMID;
			*AccessibleContext = pkg->rAccessibleContext;
			return TRUE;
		}
	}
	
	return FALSE;
}

/**
 * getAccessibleContextInfo - fills a struct with a bunch of information
 * contained in the Java Accessibility API
 *
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleContextInfo(long vmID, 
										  jobject AccessibleContext, 
										  AccessibleContextInfo *info) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleContextInfoPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleContextInfoPackage *pkg = (GetAccessibleContextInfoPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleContextInfoPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleContextInfo(%X, %X, )", vmID, AccessibleContext);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			memcpy(info, &(pkg->rAccessibleContextInfo), sizeof(AccessibleContextInfo));
			return TRUE;
		}
	}

	return FALSE;
}

/**
 * getAccessibleChildFromContext - performs the Java code:
 *   Accessible child = ac.getAccessibleChild(i);
 *   return child.getAccessibleContext();
 *
 * Note: this call explicitly goes through the AccessBridge,
 * so that the AccessBridge can hide expected changes in how this functions 
 * between JDK 1.1.x w/AccessibilityUtility classes, and JDK 1.2, when some 
 * of this functionality may be built into the platform
 *
 */
jobject
WinAccessBridge::getAccessibleChildFromContext(long vmID, 
											   jobject AccessibleContext,
											   jint childIndex) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return (jobject)0;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleChildFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleChildFromContextPackage *pkg = (GetAccessibleChildFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleChildFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
	pkg->childIndex = childIndex;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleChildFromContext(%X, %X, %d)", vmID, AccessibleContext, childIndex);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			return pkg->rAccessibleContext;
		}
	}
	
	return (jobject) 0;
}

/**
 * getAccessibleParentFromContext - returns the parent AccessibleContext jobject
 *
 * Note: this may be null, if the AccessibleContext passed in is a top-level
 * window, then it has no parent.
 *
 */
jobject
WinAccessBridge::getAccessibleParentFromContext(long vmID, 
												jobject AccessibleContext) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return (jobject)0;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleParentFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleParentFromContextPackage *pkg = (GetAccessibleParentFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleParentFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleParentFromContext(%X, %X)", vmID, AccessibleContext);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			return pkg->rAccessibleContext;
		}
	}
	
	return (jobject) 0;
}

/********** AccessibleTable routines ***********************************/

BOOL
WinAccessBridge::getAccessibleTableInfo(long vmID,
									    jobject accessibleContext,
									    AccessibleTableInfo *tableInfo) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableInfo(%X, %X, %X)", vmID, accessibleContext,
		tableInfo);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableInfoPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableInfoPackage *pkg = (GetAccessibleTableInfoPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableInfoPackage;
	pkg->vmID = vmID;
	pkg->accessibleContext = accessibleContext;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			memcpy(tableInfo, &(pkg->rTableInfo), sizeof(AccessibleTableInfo));
			if (pkg->rTableInfo.rowCount != -1) {
				PrintDebugString("  ##### WinAccessBridge::getAccessibleTableInfo succeeded");
				return TRUE;
			}
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableInfo failed");
	return FALSE;
}

BOOL 
WinAccessBridge::getAccessibleTableCellInfo(long vmID, jobject accessibleTable, 
										    jint row, jint column, 
											AccessibleTableCellInfo *tableCellInfo) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableCellInfo(%X, %X, %d, %d, %X)", vmID, 
		accessibleTable, row, column, tableCellInfo);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}

	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableCellInfoPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableCellInfoPackage *pkg = (GetAccessibleTableCellInfoPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableCellInfoPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;
	pkg->row = row;
	pkg->column = column;
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);

	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  XXXX pkg->rTableCellInfo.accessibleContext = %X", pkg->rTableCellInfo.accessibleContext);
			memcpy(tableCellInfo, &(pkg->rTableCellInfo), sizeof(AccessibleTableCellInfo));
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableCellInfo succeeded");
			return TRUE;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableCellInfo failed");
	return FALSE;
}


BOOL
WinAccessBridge::getAccessibleTableRowHeader(long vmID, jobject accessibleContext, AccessibleTableInfo *tableInfo) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableRowHeader(%X, %X, %X)", vmID, accessibleContext);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableRowHeaderPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableRowHeaderPackage *pkg = (GetAccessibleTableRowHeaderPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableRowHeaderPackage;
	pkg->vmID = vmID;
	pkg->accessibleContext = accessibleContext;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowHeader succeeded");
			memcpy(tableInfo, &(pkg->rTableInfo), sizeof(AccessibleTableInfo));
			return TRUE;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowHeader failed");
	return FALSE;
}

BOOL
WinAccessBridge::getAccessibleTableColumnHeader(long vmID, jobject accessibleContext, AccessibleTableInfo *tableInfo) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableColumnHeader(%X, %X, %X)", vmID, accessibleContext);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableColumnHeaderPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableColumnHeaderPackage *pkg = (GetAccessibleTableColumnHeaderPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableColumnHeaderPackage;
	pkg->vmID = vmID;
	pkg->accessibleContext = accessibleContext;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnHeader succeeded");
			memcpy(tableInfo, &(pkg->rTableInfo), sizeof(AccessibleTableInfo));
			return TRUE;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnHeader failed");
	return FALSE;
}

jobject
WinAccessBridge::getAccessibleTableRowDescription(long vmID,
									              jobject accessibleContext,
										          jint row) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableRowDescription(%X, %X, %d)", vmID, accessibleContext,
		row);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableRowDescriptionPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableRowDescriptionPackage *pkg = (GetAccessibleTableRowDescriptionPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableRowDescriptionPackage;
	pkg->vmID = vmID;
	pkg->row = row;
	pkg->accessibleContext = accessibleContext;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowDescription succeeded");
			return pkg->rAccessibleContext;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowDescription failed");
	return (jobject)0;
}

jobject
WinAccessBridge::getAccessibleTableColumnDescription(long vmID,
									              jobject accessibleContext,
										          jint column) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableColumnDescription(%X, %X, %d)", vmID, accessibleContext,
		column);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableColumnDescriptionPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableColumnDescriptionPackage *pkg = 
		(GetAccessibleTableColumnDescriptionPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableColumnDescriptionPackage;
	pkg->vmID = vmID;
	pkg->column = column;
	pkg->accessibleContext = accessibleContext;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnDescription succeeded");
			return pkg->rAccessibleContext;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnDescription failed");
	return (jobject)0;
}

jint 
WinAccessBridge::getAccessibleTableRowSelectionCount(long vmID, jobject accessibleTable) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableRowSelectionCount(%X, %X, %X)", vmID, accessibleTable);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return 0;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableRowSelectionCountPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableRowSelectionCountPackage *pkg = 
		(GetAccessibleTableRowSelectionCountPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableRowSelectionCountPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowSelectionCount succeeded");
			return pkg->rCount;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowSelectionCount failed");
	return 0;
}

BOOL 
WinAccessBridge::isAccessibleTableRowSelected(long vmID, jobject accessibleTable, jint row) {

	PrintDebugString("\r\n##### WinAccessBridge::isAccessibleTableRowSelected(%X, %X, %X)", vmID, accessibleTable);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(IsAccessibleTableRowSelectedPackage)];
	PackageType *type = (PackageType *) buffer;
	IsAccessibleTableRowSelectedPackage *pkg = (IsAccessibleTableRowSelectedPackage *) (buffer + sizeof(PackageType));
	*type = cIsAccessibleTableRowSelectedPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;
	pkg->row = row;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::isAccessibleTableRowSelected succeeded");
			return pkg->rResult;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::isAccessibleTableRowSelected failed");
	return FALSE;
}

BOOL 
WinAccessBridge::getAccessibleTableRowSelections(long vmID, jobject accessibleTable, jint count, jint *selections) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableRowSelections(%X, %X, %X)", vmID, accessibleTable);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableRowSelectionsPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableRowSelectionsPackage *pkg = 
		(GetAccessibleTableRowSelectionsPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableRowSelectionsPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;
	pkg->count = count;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowSelections succeeded");
			memcpy(selections, pkg->rSelections, count * sizeof(jint));
			return TRUE;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRowSelections failed");
	return FALSE;
}


jint 
WinAccessBridge::getAccessibleTableColumnSelectionCount(long vmID, jobject accessibleTable) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableColumnSelectionCount(%X, %X, %X)", vmID, 
		accessibleTable);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableColumnSelectionCountPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableColumnSelectionCountPackage *pkg = 
		(GetAccessibleTableColumnSelectionCountPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableColumnSelectionCountPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnSelectionCount succeeded");
			return pkg->rCount;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnSelectionCount failed");
	return 0;
}

BOOL 
WinAccessBridge::isAccessibleTableColumnSelected(long vmID, jobject accessibleTable, jint column) {
	PrintDebugString("\r\n##### WinAccessBridge::isAccessibleTableColumnSelected(%X, %X, %X)", vmID, accessibleTable);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(IsAccessibleTableColumnSelectedPackage)];
	PackageType *type = (PackageType *) buffer;
	IsAccessibleTableColumnSelectedPackage *pkg = (IsAccessibleTableColumnSelectedPackage *) (buffer + sizeof(PackageType));
	*type = cIsAccessibleTableColumnSelectedPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;
	pkg->column = column;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::isAccessibleTableColumnSelected succeeded");
			return pkg->rResult;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::isAccessibleTableColumnSelected failed");
	return FALSE;
}

BOOL 
WinAccessBridge::getAccessibleTableColumnSelections(long vmID, jobject accessibleTable, jint count, 
													jint *selections) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableColumnSelections(%X, %X, %X)", vmID, accessibleTable);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableColumnSelectionsPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableColumnSelectionsPackage *pkg = 
		(GetAccessibleTableColumnSelectionsPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableColumnSelectionsPackage;
	pkg->vmID = vmID;
	pkg->count = count;
	pkg->accessibleTable = accessibleTable;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnSelections succeeded");
			memcpy(selections, pkg->rSelections, count * sizeof(jint));
			return TRUE;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumnSelections failed");
	return FALSE;
}

jint 
WinAccessBridge::getAccessibleTableRow(long vmID, jobject accessibleTable, jint index) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableRow(%X, %X, index=%d)", vmID, 
		accessibleTable, index);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableRowPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableRowPackage *pkg = 
		(GetAccessibleTableRowPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableRowPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;
	pkg->index = index;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRow succeeded");
			return pkg->rRow;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableRow failed");
	return 0;
}

jint 
WinAccessBridge::getAccessibleTableColumn(long vmID, jobject accessibleTable, jint index) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableColumn(%X, %X, index=%d)", vmID, 
		accessibleTable, index);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableColumnPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableColumnPackage *pkg = 
		(GetAccessibleTableColumnPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableColumnPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;
	pkg->index = index;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumn succeeded");
			return pkg->rColumn;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableColumn failed");
	return 0;
}

jint 
WinAccessBridge::getAccessibleTableIndex(long vmID, jobject accessibleTable, jint row, jint column) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleTableIndex(%X, %X, row=%d, col=%d)", vmID, 
		accessibleTable, row, column);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTableIndexPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTableIndexPackage *pkg = 
		(GetAccessibleTableIndexPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTableIndexPackage;
	pkg->vmID = vmID;
	pkg->accessibleTable = accessibleTable;
	pkg->row = row;
	pkg->column = column;

 	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### WinAccessBridge::getAccessibleTableIndex succeeded");
			return pkg->rIndex;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleTableIndex failed");
	return 0;
}

/********** end AccessibleTable routines ******************************/

BOOL 
WinAccessBridge::getAccessibleRelationSet(long vmID, jobject accessibleContext, 
										  AccessibleRelationSetInfo *relationSetInfo) {

	PrintDebugString("\r\n##### WinAccessBridge::getAccessibleRelationSet(%X, %X, %X)", vmID, 
		accessibleContext, relationSetInfo);

	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}

	char buffer[sizeof(PackageType) + sizeof(GetAccessibleRelationSetPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleRelationSetPackage *pkg = (GetAccessibleRelationSetPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleRelationSetPackage;
	pkg->vmID = vmID;
	pkg->accessibleContext = accessibleContext;

	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			PrintDebugString("  ##### pkg->rAccessibleRelationSetInfo.relationCount = %X", 
				pkg->rAccessibleRelationSetInfo.relationCount);
			memcpy(relationSetInfo, &(pkg->rAccessibleRelationSetInfo), sizeof(AccessibleRelationSetInfo));
			PrintDebugString("  ##### WinAccessBridge::getAccessibleRelationSet succeeded");
			return TRUE;
		}
	}
	PrintDebugString("  ##### WinAccessBridge::getAccessibleRelationSet failed");
	return FALSE;
}

/********** AccessibleText routines ***********************************/

/**
 * getAccessibleTextInfo - fills a struct with a bunch of information
 * contained in the Java Accessibility AccessibleText API
 *
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleTextInfo(long vmID,
									   jobject AccessibleContext,
									   AccessibleTextInfo *textInfo, 
									   jint x, jint y) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTextInfoPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTextInfoPackage *pkg = (GetAccessibleTextInfoPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTextInfoPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
	pkg->x = x;
	pkg->y = y;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleTextInfo(%X, %X, %X, %d, %d)", vmID, AccessibleContext, textInfo, x, y);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			memcpy(textInfo, &(pkg->rTextInfo), sizeof(AccessibleTextInfo));
			if (pkg->rTextInfo.charCount != -1) {
				return TRUE;
			}
		}
	}

	return FALSE;
}

/**
 * getAccessibleTextItems - fills a struct with letter, word, and sentence info
 * of the AccessibleText interface at a given index
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleTextItems(long vmID, 
										jobject AccessibleContext,
										AccessibleTextItemsInfo *textItems,
										jint index) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTextItemsPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTextItemsPackage *pkg = (GetAccessibleTextItemsPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTextItemsPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
	pkg->index = index;
		// zero things out, in case the call fails
	pkg->rTextItemsInfo.letter = '\0';
	pkg->rTextItemsInfo.word[0] = '\0';
	pkg->rTextItemsInfo.sentence[0] = '\0';

    PrintDebugString("\r\nWinAccessBridge::getAccessibleTextItems(%X, %X, %X, %d)", vmID, AccessibleContext, textItems, index);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			memcpy(textItems, &(pkg->rTextItemsInfo), sizeof(AccessibleTextItemsInfo));
			if (pkg->rTextItemsInfo.letter != '/0') {
				return TRUE;
			}
		}
	}

	return FALSE;
}

/**
 * getAccessibleTextSelectionInfo - returns information about the selected
 * text of the object implementing AccessibleText
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleTextSelectionInfo(long vmID,
												jobject AccessibleContext,
												AccessibleTextSelectionInfo *selectionInfo) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTextSelectionInfoPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTextSelectionInfoPackage *pkg = (GetAccessibleTextSelectionInfoPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTextSelectionInfoPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleTextSelectionInfo(%X, %X, %X)", vmID, AccessibleContext, selectionInfo);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			memcpy(selectionInfo, &(pkg->rTextSelectionItemsInfo), sizeof(AccessibleTextSelectionInfo));
			// [[[FIXME]]] should test to see if valid info returned; return FALSE if not
			return TRUE;
		}
	}

	return FALSE;
}

/**
 * getAccessibleTextAttributes - performs the Java code:
 *   ...[[[FIXME]]] fill in this comment...
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleTextAttributes(long vmID,
											 jobject AccessibleContext,
											 jint index,
                                             AccessibleTextAttributesInfo *attributes) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTextAttributeInfoPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTextAttributeInfoPackage *pkg = (GetAccessibleTextAttributeInfoPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTextAttributeInfoPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
	pkg->index = index;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleTextAttributes(%X, %X, %d, %X)", vmID, AccessibleContext, index, attributes);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			memcpy(attributes, &(pkg->rAttributeInfo), sizeof(AccessibleTextAttributesInfo));
			return TRUE;
		}
	}

	return FALSE;
}

/**
 * getAccessibleTextRect - performs the Java code:
 *   ...[[[FIXME]]] fill in this comment...
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleTextRect(long vmID,
									   jobject AccessibleContext,
									   AccessibleTextRectInfo *rectInfo,
									   jint index) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTextRectInfoPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTextRectInfoPackage *pkg = (GetAccessibleTextRectInfoPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTextRectInfoPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
	pkg->index = index;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleTextRect(%X, %X, %X, %d)", vmID, AccessibleContext, rectInfo, index);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			memcpy(rectInfo, (&pkg->rTextRectInfo), sizeof(AccessibleTextRectInfo));
			// [[[FIXME]]] should test to see if valid info returned; return FALSE if not
			return TRUE;
		}
	}

	return FALSE;
}


/**
 * getAccessibleTextLineBounds - performs the Java code:
 *   ...[[[FIXME]]] fill in this comment...
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleTextLineBounds(long vmID,
									   jobject AccessibleContext,
									   jint index, jint *startIndex, jint *endIndex) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTextLineBoundsPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTextLineBoundsPackage *pkg = (GetAccessibleTextLineBoundsPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTextLineBoundsPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
	pkg->index = index;

    PrintDebugString("\r\nWinAccessBridge::getAccessibleTextLineBounds(%X, %X, %d, )", vmID, AccessibleContext, index);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			*startIndex = pkg->rLineStart;
			*endIndex = pkg->rLineEnd;
			// [[[FIXME]]] should test to see if valid info returned; return FALSE if not
			return TRUE;
		}
	}

	return FALSE;
}


/**
 * getAccessibleTextLineBounds - performs the Java code:
 *   ...[[[FIXME]]] fill in this comment...
 *
 * Note: if the AccessibleContext parameter is bogus, this call will blow up
 */
BOOL
WinAccessBridge::getAccessibleTextRange(long vmID,
									   jobject AccessibleContext,
									   jint start, jint end, wchar_t *text, short len) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleTextRangePackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleTextRangePackage *pkg = (GetAccessibleTextRangePackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleTextRangePackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
	pkg->start = start;
	pkg->end = end;


    PrintDebugString("\r\nWinAccessBridge::getAccessibleTextRange(%X, %X, %d, %d, )", vmID, AccessibleContext, start, end);
	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			wcsncpy(text, pkg->rText, len);
			// [[[FIXME]]] should test to see if valid info returned; return FALSE if not
			return TRUE;
		}
	}

	return FALSE;
}




/********** AccessibleValue routines ***************/

BOOL
WinAccessBridge::getCurrentAccessibleValueFromContext(long vmID,
                                                      jobject AccessibleContext,
                                                      wchar_t *value, short len) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetCurrentAccessibleValueFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	GetCurrentAccessibleValueFromContextPackage *pkg = (GetCurrentAccessibleValueFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cGetCurrentAccessibleValueFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			wcsncpy(value, pkg->rValue, len);
			// [[[FIXME]]] should test to see if valid info returned; return FALSE if not
			return TRUE;
		}
	}

	return FALSE;
}

BOOL
WinAccessBridge::getMaximumAccessibleValueFromContext(long vmID,
                                                      jobject AccessibleContext,
                                                      wchar_t *value, short len) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetMaximumAccessibleValueFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	GetMaximumAccessibleValueFromContextPackage *pkg = (GetMaximumAccessibleValueFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cGetMaximumAccessibleValueFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			wcsncpy(value, pkg->rValue, len);
			// [[[FIXME]]] should test to see if valid info returned; return FALSE if not
			return TRUE;
		}
	}

	return FALSE;
}

BOOL
WinAccessBridge::getMinimumAccessibleValueFromContext(long vmID,
                                                      jobject AccessibleContext,
                                                      wchar_t *value, short len) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(GetMinimumAccessibleValueFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	GetMinimumAccessibleValueFromContextPackage *pkg = (GetMinimumAccessibleValueFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cGetMinimumAccessibleValueFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
			wcsncpy(value, pkg->rValue, len);
			// [[[FIXME]]] should test to see if valid info returned; return FALSE if not
			return TRUE;
		}
	}

	return FALSE;
}


/********** AccessibleSelection routines ***************/

void
WinAccessBridge::addAccessibleSelectionFromContext(long vmID,
                                                   jobject AccessibleContext, int i) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(AddAccessibleSelectionFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	AddAccessibleSelectionFromContextPackage *pkg = (AddAccessibleSelectionFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cAddAccessibleSelectionFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
    pkg->index = i;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		sendMemoryPackage(buffer, sizeof(buffer), destABWindow);
	}
}

void
WinAccessBridge::clearAccessibleSelectionFromContext(long vmID,
                                                     jobject AccessibleContext) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(ClearAccessibleSelectionFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	ClearAccessibleSelectionFromContextPackage *pkg = (ClearAccessibleSelectionFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cClearAccessibleSelectionFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		sendMemoryPackage(buffer, sizeof(buffer), destABWindow);
	}
}

jobject
WinAccessBridge::getAccessibleSelectionFromContext(long vmID,
                                                   jobject AccessibleContext, int i) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return (jobject)0;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleSelectionFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleSelectionFromContextPackage *pkg = (GetAccessibleSelectionFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleSelectionFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
    pkg->index = i;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
            return pkg->rAccessibleContext;
        }
	}

    return (jobject) 0;
}

int
WinAccessBridge::getAccessibleSelectionCountFromContext(long vmID,
                                                        jobject AccessibleContext) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return -1;
	}
	char buffer[sizeof(PackageType) + sizeof(GetAccessibleSelectionCountFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	GetAccessibleSelectionCountFromContextPackage *pkg = (GetAccessibleSelectionCountFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cGetAccessibleSelectionCountFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
            return (int) pkg->rCount;
        }
	}

    return -1;
}

BOOL
WinAccessBridge::isAccessibleChildSelectedFromContext(long vmID,
                                                      jobject AccessibleContext, int i) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return FALSE;
	}
	char buffer[sizeof(PackageType) + sizeof(IsAccessibleChildSelectedFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	IsAccessibleChildSelectedFromContextPackage *pkg = (IsAccessibleChildSelectedFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cIsAccessibleChildSelectedFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
    pkg->index = i;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		if (sendMemoryPackage(buffer, sizeof(buffer), destABWindow) == TRUE) {
            if (pkg->rResult != 0) {
                return TRUE;
            }
        }
	}

    return FALSE;
}


void
WinAccessBridge::removeAccessibleSelectionFromContext(long vmID,
                                                      jobject AccessibleContext, int i) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(RemoveAccessibleSelectionFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	RemoveAccessibleSelectionFromContextPackage *pkg = (RemoveAccessibleSelectionFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cRemoveAccessibleSelectionFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;
    pkg->index = i;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		sendMemoryPackage(buffer, sizeof(buffer), destABWindow);
	}
}

void
WinAccessBridge::selectAllAccessibleSelectionFromContext(long vmID,
                                                         jobject AccessibleContext) {
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(SelectAllAccessibleSelectionFromContextPackage)];
	PackageType *type = (PackageType *) buffer;
	SelectAllAccessibleSelectionFromContextPackage *pkg = (SelectAllAccessibleSelectionFromContextPackage *) (buffer + sizeof(PackageType));
	*type = cSelectAllAccessibleSelectionFromContextPackage;
	pkg->vmID = vmID;
	pkg->AccessibleContext = AccessibleContext;

	// need to call only the HWND/VM that contains this AC
	HWND destABWindow = javaVMs->findAccessBridgeWindow(vmID);
	if (destABWindow != (HWND) 0) {
		sendMemoryPackage(buffer, sizeof(buffer), destABWindow);
	}
}


/*********** Event handling methods **********************************/

/**
 * addEventNotification - tell all Java-launched AccessBridge DLLs
 *                        that we want events of the specified type
 *
 * [[[FIXME]]] since we're just sending a long & a source window,
 *			   we could use a private message rather than WM_COPYDATA
 *			   (though we still may want it to be synchronous; dunno...)
 *
 */
void
WinAccessBridge::addJavaEventNotification(jlong type) {
	PrintDebugString("\r\nWinAccessBridge::addJavaEventNotification(%X)", type);
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}

	char buffer[sizeof(PackageType) + sizeof(AddJavaEventNotificationPackage)];
	PackageType *pkgType = (PackageType *) buffer;
	AddJavaEventNotificationPackage *pkg = (AddJavaEventNotificationPackage *) (buffer + sizeof(PackageType));
	*pkgType = cAddJavaEventNotificationPackage;
	pkg->type = type;
	pkg->DLLwindow = dialogWindow;

	PrintDebugString("  ->pkgType = , eventType = , DLLwindow = ", 
		  *pkgType, pkg->type, pkg->DLLwindow);

		// send addEventNotification message to all JVMs
	AccessBridgeJavaVMInstance *current = javaVMs;
	while (current != (AccessBridgeJavaVMInstance *) 0) {
		current->sendPackage(buffer, sizeof(buffer));		// no return values!
		current = current->nextJVMInstance;
	}
}

/**
 * removeEventNotification - tell all Java-launched AccessBridge DLLs
 *							 that we no longer want events of the 
 *							 specified type
 *
 * [[[FIXME]]] since we're just sending a long & a source window,
 *			   we could use a private message rather than WM_COPYDATA
 *			   (though we still may want it to be synchronous; dunno...)
 *
 */
void
WinAccessBridge::removeJavaEventNotification(jlong type) {
	PrintDebugString("\r\nin WinAccessBridge::removeJavaEventNotification(%X)", type);
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(RemoveJavaEventNotificationPackage)];
	PackageType *pkgType = (PackageType *) buffer;
	RemoveJavaEventNotificationPackage *pkg = (RemoveJavaEventNotificationPackage *) (buffer + sizeof(PackageType));
	*pkgType = cRemoveJavaEventNotificationPackage;
	pkg->type = type;
	pkg->DLLwindow = dialogWindow;

	PrintDebugString("  ->pkgType = , eventType = , DLLwindow = ", 
		  *pkgType, pkg->type, pkg->DLLwindow);

		// send removeEventNotification message to all JVMs
	AccessBridgeJavaVMInstance *current = javaVMs;
	while (current != (AccessBridgeJavaVMInstance *) 0) {
		current->sendPackage(buffer, sizeof(buffer));		// no return values!
		current = current->nextJVMInstance;
	}
}


/*********** Event handling methods **********************************/

/**
 * addAccessibilityEventNotification - tell all Java-launched AccessBridge DLLs
 *                        that we want events of the specified type
 *
 * [[[FIXME]]] since we're just sending a long & a source window,
 *			   we could use a private message rather than WM_COPYDATA
 *			   (though we still may want it to be synchronous; dunno...)
 *
 */
void
WinAccessBridge::addAccessibilityEventNotification(jlong type) {
	PrintDebugString("\r\nin WinAccessBridge::addAccessibilityEventNotification(%X)", type);
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(AddAccessibilityEventNotificationPackage)];
	PackageType *pkgType = (PackageType *) buffer;
	AddAccessibilityEventNotificationPackage *pkg = (AddAccessibilityEventNotificationPackage *) (buffer + sizeof(PackageType));
	*pkgType = cAddAccessibilityEventNotificationPackage;
	pkg->type = type;
	pkg->DLLwindow = dialogWindow;

	PrintDebugString("  ->pkgType = , eventType = , DLLwindow = ", 
		  *pkgType, pkg->type, pkg->DLLwindow);

		// send addEventNotification message to all JVMs
	AccessBridgeJavaVMInstance *current = javaVMs;
	while (current != (AccessBridgeJavaVMInstance *) 0) {
		current->sendPackage(buffer, sizeof(buffer));		// no return values!
		current = current->nextJVMInstance;
	}
}

/**
 * removeAccessibilityEventNotification - tell all Java-launched AccessBridge DLLs
 *							 that we no longer want events of the 
 *							 specified type
 *
 * [[[FIXME]]] since we're just sending a long & a source window,
 *			   we could use a private message rather than WM_COPYDATA
 *			   (though we still may want it to be synchronous; dunno...)
 *
 */
void
WinAccessBridge::removeAccessibilityEventNotification(jlong type) {
	PrintDebugString("\r\nin WinAccessBridge::removeAccessibilityEventNotification(%X)", type);
	if ((AccessBridgeJavaVMInstance *) 0 == javaVMs) {
		return;
	}
	char buffer[sizeof(PackageType) + sizeof(RemoveAccessibilityEventNotificationPackage)];
	PackageType *pkgType = (PackageType *) buffer;
	RemoveAccessibilityEventNotificationPackage *pkg = (RemoveAccessibilityEventNotificationPackage *) (buffer + sizeof(PackageType));
	*pkgType = cRemoveAccessibilityEventNotificationPackage;
	pkg->type = type;
	pkg->DLLwindow = dialogWindow;

	PrintDebugString("  ->pkgType = , eventType = , DLLwindow = ", 
		  *pkgType, pkg->type, pkg->DLLwindow);

		// send removeEventNotification message to all JVMs
	AccessBridgeJavaVMInstance *current = javaVMs;
	while (current != (AccessBridgeJavaVMInstance *) 0) {
		current->sendPackage(buffer, sizeof(buffer));		// no return values!
		current = current->nextJVMInstance;
	}
}


#define CALL_SET_EVENT_FP(function, callbackFP)		\
	void WinAccessBridge::function(callbackFP fp) {	\
		eventHandler->function(fp, this);			\
		/* eventHandler calls back to winAccessBridgeDLL to set eventMask */	\
	}

CALL_SET_EVENT_FP(setPropertyChangeFP, AccessBridge_PropertyChangeFP)
CALL_SET_EVENT_FP(setFocusGainedFP, AccessBridge_FocusGainedFP)
CALL_SET_EVENT_FP(setFocusLostFP, AccessBridge_FocusLostFP)
CALL_SET_EVENT_FP(setCaretUpdateFP, AccessBridge_CaretUpdateFP)
CALL_SET_EVENT_FP(setMouseClickedFP, AccessBridge_MouseClickedFP)
CALL_SET_EVENT_FP(setMouseEnteredFP, AccessBridge_MouseEnteredFP)
CALL_SET_EVENT_FP(setMouseExitedFP, AccessBridge_MouseExitedFP)
CALL_SET_EVENT_FP(setMousePressedFP, AccessBridge_MousePressedFP)
CALL_SET_EVENT_FP(setMouseReleasedFP, AccessBridge_MouseReleasedFP)
CALL_SET_EVENT_FP(setMenuCanceledFP, AccessBridge_MenuCanceledFP)
CALL_SET_EVENT_FP(setMenuDeselectedFP, AccessBridge_MenuDeselectedFP)
CALL_SET_EVENT_FP(setMenuSelectedFP, AccessBridge_MenuSelectedFP)
CALL_SET_EVENT_FP(setPopupMenuCanceledFP, AccessBridge_PopupMenuCanceledFP)
CALL_SET_EVENT_FP(setPopupMenuWillBecomeInvisibleFP, AccessBridge_PopupMenuWillBecomeInvisibleFP)
CALL_SET_EVENT_FP(setPopupMenuWillBecomeVisibleFP, AccessBridge_PopupMenuWillBecomeVisibleFP)

CALL_SET_EVENT_FP(setPropertyNameChangeFP, AccessBridge_PropertyNameChangeFP)
CALL_SET_EVENT_FP(setPropertyDescriptionChangeFP, AccessBridge_PropertyDescriptionChangeFP)
CALL_SET_EVENT_FP(setPropertyStateChangeFP, AccessBridge_PropertyStateChangeFP)
CALL_SET_EVENT_FP(setPropertyValueChangeFP, AccessBridge_PropertyValueChangeFP)
CALL_SET_EVENT_FP(setPropertySelectionChangeFP, AccessBridge_PropertySelectionChangeFP)
CALL_SET_EVENT_FP(setPropertyTextChangeFP, AccessBridge_PropertyTextChangeFP)
CALL_SET_EVENT_FP(setPropertyCaretChangeFP, AccessBridge_PropertyCaretChangeFP)
CALL_SET_EVENT_FP(setPropertyVisibleDataChangeFP, AccessBridge_PropertyVisibleDataChangeFP)
CALL_SET_EVENT_FP(setPropertyChildChangeFP, AccessBridge_PropertyChildChangeFP)
CALL_SET_EVENT_FP(setPropertyActiveDescendentChangeFP, AccessBridge_PropertyActiveDescendentChangeFP)

CALL_SET_EVENT_FP(setPropertyTableModelChangeFP, AccessBridge_PropertyTableModelChangeFP)
