/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * @(#)WinAccessBridge.h	1.11 03/09/02
 */

/* 
 * A DLL which is loaded by Windows executables to handle communication
 * between Java VMs purposes of Accessbility.
 */

#ifndef __WinAccessBridge_H__
#define __WinAccessBridge_H__

#include <windows.h>
#include "AccessBridgePackages.h"
#include "AccessBridgeEventHandler.h"
#include "AccessBridgeJavaVMInstance.h"
#include "AccessBridgeMessageQueue.h"


extern "C" {
	BOOL WINAPI DllMain(HINSTANCE hinstDll, DWORD fdwReason, 
						LPVOID lpvReserved);
	void AppendToCallOutput(char *s);
	BOOL CALLBACK AccessBridgeDialogProc(HWND hDlg, UINT message, 
										 UINT wParam, LONG lParam);
	HWND getTopLevelHWND(HWND descendant);
}


/**
 * The WinAccessBridge class.  The core of the Windows AT AccessBridge dll
 */
class WinAccessBridge {
	HINSTANCE windowsInstance;
	HWND dialogWindow;
	AccessBridgeJavaVMInstance *javaVMs;
	AccessBridgeEventHandler *eventHandler;
	AccessBridgeMessageQueue *messageQueue;

public:
	WinAccessBridge(HINSTANCE hInstance);
	~WinAccessBridge();
	BOOL initWindow();

	// IPC with the Java AccessBridge DLL
	LRESULT rendezvousWithNewJavaDLL(HWND JavaBridgeDLLwindow, long vmID);

	void sendPackage(char *buffer, long bufsize, HWND destWindow);
	BOOL sendMemoryPackage(char *buffer, long bufsize, HWND destWindow);
	BOOL queuePackage(char *buffer, long bufsize);
	BOOL receiveAQueuedPackage();
	void preProcessPackage(char *buffer, long bufsize);
	void processPackage(char *buffer, long bufsize);
	void JavaVMDestroyed(HWND VMBridgeDLLWindow);

	// Java VM object memory management
	void releaseJavaObject(long vmID, jobject object);

	// Version info
    BOOL getVersionInfo(long vmID, AccessBridgeVersionInfo *info);

	// HWND management methods
	HWND getNextJavaWindow(HWND previous);
	BOOL isJavaWindow(HWND window);
	BOOL isSameObject(long vmID, jobject obj1, jobject obj2);
	BOOL getAccessibleContextFromHWND(HWND window, long *vmID, jobject *AccessibleContext);

	// Accessible Context methods
	BOOL getAccessibleContextAt(long vmID, jobject AccessibleContextParent, 
								jint x, jint y, jobject *AccessibleContext);
	BOOL getAccessibleContextWithFocus(HWND window, long *vmID, jobject *AccessibleContext);
	BOOL getAccessibleContextInfo(long vmID, jobject AccessibleContext, AccessibleContextInfo *info);
	jobject getAccessibleChildFromContext(long vmID, jobject AccessibleContext, jint childIndex);
	jobject getAccessibleParentFromContext(long vmID, jobject AccessibleContext);

	/* begin AccessibleTable methods */
	BOOL getAccessibleTableInfo(long vmID, jobject acParent, AccessibleTableInfo *tableInfo);
	BOOL getAccessibleTableCellInfo(long vmID, jobject accessibleTable, jint row, jint column, 
		AccessibleTableCellInfo *tableCellInfo);

	BOOL getAccessibleTableRowHeader(long vmID, jobject acParent, AccessibleTableInfo *tableInfo);
	BOOL getAccessibleTableColumnHeader(long vmID, jobject acParent, AccessibleTableInfo *tableInfo);

	jobject getAccessibleTableRowDescription(long vmID, jobject acParent, jint row);
	jobject getAccessibleTableColumnDescription(long vmID, jobject acParent, jint column);

	jint getAccessibleTableRowSelectionCount(long vmID, jobject accessibleTable);
	BOOL isAccessibleTableRowSelected(long vmID, jobject accessibleTable, jint row);
	BOOL getAccessibleTableRowSelections(long vmID, jobject accessibleTable, jint count, 
		jint *selections);

	jint getAccessibleTableColumnSelectionCount(long vmID, jobject accessibleTable);
	BOOL isAccessibleTableColumnSelected(long vmID, jobject accessibleTable, jint column);
	BOOL getAccessibleTableColumnSelections(long vmID, jobject accessibleTable, jint count, 
		jint *selections);

	jint getAccessibleTableRow(long vmID, jobject accessibleTable, jint index);
	jint getAccessibleTableColumn(long vmID, jobject accessibleTable, jint index);
	jint getAccessibleTableIndex(long vmID, jobject accessibleTable, jint row, jint column);

	/* end AccessibleTable methods */

	BOOL getAccessibleRelationSet(long vmID, jobject accessibleContext,
		AccessibleRelationSetInfo *relationSetInfo);


	// Accessible Text methods
	BOOL getAccessibleTextInfo(long vmID, jobject AccessibleContext, AccessibleTextInfo *textInfo, jint x, jint y);
	BOOL getAccessibleTextItems(long vmID, jobject AccessibleContext, AccessibleTextItemsInfo *textItems, jint index);
	BOOL getAccessibleTextSelectionInfo(long vmID, jobject AccessibleContext, AccessibleTextSelectionInfo *selectionInfo);
	BOOL getAccessibleTextAttributes(long vmID, jobject AccessibleContext, jint index, AccessibleTextAttributesInfo *attributes);
	BOOL getAccessibleTextRect(long vmID, jobject AccessibleContext, AccessibleTextRectInfo *rectInfo, jint index);
    BOOL getAccessibleTextLineBounds(long vmID, jobject AccessibleContext, jint index, jint *startIndex, jint *endIndex);
    BOOL getAccessibleTextRange(long vmID, jobject AccessibleContext, jint start, jint end, wchar_t *text, short len);

    // Accessible Value methods
    BOOL getCurrentAccessibleValueFromContext(long vmID, jobject AccessibleContext, wchar_t *value, short len);
    BOOL getMaximumAccessibleValueFromContext(long vmID, jobject AccessibleContext, wchar_t *value, short len);
    BOOL getMinimumAccessibleValueFromContext(long vmID, jobject AccessibleContext, wchar_t *value, short len);

    // Accessible Selection methods
    void addAccessibleSelectionFromContext(long vmID, jobject AccessibleContext, int i);
    void clearAccessibleSelectionFromContext(long vmID, jobject AccessibleContext);
    jobject getAccessibleSelectionFromContext(long vmID, jobject AccessibleContext, int i);
    int getAccessibleSelectionCountFromContext(long vmID, jobject AccessibleContext);
    BOOL isAccessibleChildSelectedFromContext(long vmID, jobject AccessibleContext, int i);
    void removeAccessibleSelectionFromContext(long vmID, jobject AccessibleContext, int i);
    void selectAllAccessibleSelectionFromContext(long vmID, jobject AccessibleContext);

    // Event handling methods
	void addJavaEventNotification(jlong type);
	void removeJavaEventNotification(jlong type);
	void addAccessibilityEventNotification(jlong type);
	void removeAccessibilityEventNotification(jlong type);

	void setPropertyChangeFP(AccessBridge_PropertyChangeFP fp);
	void setFocusGainedFP(AccessBridge_FocusGainedFP fp);
	void setFocusLostFP(AccessBridge_FocusLostFP fp);
	void setCaretUpdateFP(AccessBridge_CaretUpdateFP fp);
	void setMouseClickedFP(AccessBridge_MouseClickedFP fp);
	void setMouseEnteredFP(AccessBridge_MouseEnteredFP fp);
	void setMouseExitedFP(AccessBridge_MouseExitedFP fp);
	void setMousePressedFP(AccessBridge_MousePressedFP fp);
	void setMouseReleasedFP(AccessBridge_MouseReleasedFP fp);
	void setMenuCanceledFP(AccessBridge_MenuCanceledFP fp);
	void setMenuDeselectedFP(AccessBridge_MenuDeselectedFP fp);
	void setMenuSelectedFP(AccessBridge_MenuSelectedFP fp);
	void setPopupMenuCanceledFP(AccessBridge_PopupMenuCanceledFP fp);
	void setPopupMenuWillBecomeInvisibleFP(AccessBridge_PopupMenuWillBecomeInvisibleFP fp);
	void setPopupMenuWillBecomeVisibleFP(AccessBridge_PopupMenuWillBecomeVisibleFP fp);

	void setPropertyNameChangeFP(AccessBridge_PropertyNameChangeFP fp);
	void setPropertyDescriptionChangeFP(AccessBridge_PropertyDescriptionChangeFP fp);
	void setPropertyStateChangeFP(AccessBridge_PropertyStateChangeFP fp);
	void setPropertyValueChangeFP(AccessBridge_PropertyValueChangeFP fp);
	void setPropertySelectionChangeFP(AccessBridge_PropertySelectionChangeFP fp);
	void setPropertyTextChangeFP(AccessBridge_PropertyTextChangeFP fp);
	void setPropertyCaretChangeFP(AccessBridge_PropertyCaretChangeFP fp);
	void setPropertyVisibleDataChangeFP(AccessBridge_PropertyVisibleDataChangeFP fp);
	void setPropertyChildChangeFP(AccessBridge_PropertyChildChangeFP fp);
	void setPropertyActiveDescendentChangeFP(AccessBridge_PropertyActiveDescendentChangeFP fp);

	void setPropertyTableModelChangeFP(AccessBridge_PropertyTableModelChangeFP fp);

};

#endif

